class ImagingSetupParameters:

    ti0 = None
    ni0 = None
    ni = None
    tg0 = None
    tg = None
    ng0 = None
    ng = None
    ns = None
    wavelength = None
    k0 = None
    M = None
    NA = None
    alpha = None
    pixel_size_physical = None
